import numpy as np
from scipy.integrate import simpson

from collapse_stats import PressSchechterModel, PeaksTheoryModel, NonLinearModel

class MassFunction:
    #def __init__(self, power_spectrum, mass_variance, statistics, K=10, C_c=0.25, gamma=0.36):
     #3   """
       # statistics: an instance of a statistics model class (e.g. PressSchechterModel),
        #"""
        #self.statistics = statistics

    def __init__(self, mass_variance, statistics, **kwargs):
        self.mass_variance = mass_variance
        self.power_spectrum = mass_variance.power_spectrum

        # statistics model can be a string or a model instance
        if isinstance(statistics, str):
            statistics = statistics.lower()
            model_map = {
                'press': PressSchechterModel,
                'peaks': PeaksTheoryModel,
                'nonlinear': NonLinearModel,
            }
        if isinstance(statistics, str):
            stats_key = statistics.lower()
            if stats_key != mass_variance.statistics.lower():
                raise ValueError(f"Inconsistent 'statistics': MassVariance has '{mass_variance.statistics}', "
                                f"but MassFunction received '{statistics}'.")
            if stats_key not in model_map:
                raise ValueError(f"Unknown statistics model '{statistics}'. Must be one of {list(model_map)}")
            self.statistics = model_map[stats_key](
                power_spectrum=self.power_spectrum,
                mass_variance=mass_variance,
                **kwargs
            )

        else:
            self.statistics = statistics
            # Optionally update its parameters with kwargs if it's already an object
            for key, value in kwargs.items():
                if hasattr(self.statistics, key):
                    setattr(self.statistics, key, value)

    def evaluate(self, k_values, mpbh_vals=50):
        """
        Compute the mass function f(M) by passing k_values to the statistics model.

        Parameters
        ----------
        k_values : np.ndarray
            Array of k modes used to define the power spectrum and smoothing.

        Returns
        -------
        m : np.ndarray
            PBH masses.
        f_m : np.ndarray
            Corresponding PBH mass function f(M).
        """
        return self.statistics.evaluate(k_values, mpbh_vals)

    def fpbh(self, fofm, m):
        m, fofm = m[~np.isnan(fofm)], fofm[~np.isnan(fofm)]
        return simpson(fofm/m, m)



import numpy as np
from scipy.interpolate import interp1d

class PowerSpectrum:
    def __init__(self,
                 shape: str = None,
                 amplitude: float = None,
                 k_star: float = None,
                 sigma_ln: float = None,
                 k_min: float = None,
                 k_max: float = None,
                 ng: float = None,
                 nd: float = None,
                 min_amplitude: float = 2e-9,
                 k_values: np.ndarray = None,
                 P_k_values: np.ndarray = None):
        """
        Parameters:
        - shape: str, type of power spectrum ('delta', 'flat', 'piecewise', or 'custom')
        - amplitude: float, overall normalization A
        - k_star: float, peak or transition scale
        - k_min, k_max: float, used in flat spectrum
        - ng: float, slope for k < k_star in piecewise shape
        - nd: float, slope for k > k_star in piecewise shape
        - min_amplitude: float, minimum cutoff to avoid zero power
        - k_values, P_k_values: optional arrays to define a custom power spectrum
        """
        self.shape = shape
        self.amplitude = amplitude
        self.k_star = k_star
        self.sigma_ln = sigma_ln
        self.k_min = k_min
        self.k_max = k_max
        self.ng = ng
        self.nd = nd
        self.min_amplitude = min_amplitude

        if k_values is not None and P_k_values is not None:
            if len(k_values) != len(P_k_values):
                raise ValueError("k_values and P_k_values must have the same length")
            self.k_values = k_values
            self.P_k_values = P_k_values
            self.interpolator = interp1d(np.log(k_values), np.log(P_k_values),
                                         bounds_error=False, fill_value=-np.inf)
        else:
            self.k_values = None
            self.P_k_values = None
            self.interpolator = None

    def __call__(self, k: np.ndarray) -> np.ndarray:
        if self.shape == "custom":
            if self.interpolator is None:
                raise ValueError("Interpolator is not defined for custom shape")
            return np.exp(self.interpolator(np.log(k)))

        return self.generate_P_k(k)

    def generate_P_k(self, k: np.ndarray) -> np.ndarray:
        A = self.amplitude
        k_star = self.k_star
        k_min = self.k_min
        k_max = self.k_max

        if self.shape == "delta":
            # δ(k - k_star) approximation using a narrow log-normal bump
            if None in (A, k_star):
                raise ValueError("Amplitude and k_star must be specified for flat shape")
            sigma = 0.001
            return A * np.exp(-0.5 * (np.log(k / k_star) / sigma) ** 2)

        elif self.shape == "flat":
            # Top-hat in log-space
            if None in (A, k_min, k_max):
                raise ValueError("Amplitude, k_min and k_max must be specified for flat shape")
            mask = (k >= k_min) & (k <= k_max)
            return A * mask.astype(float)

        elif self.shape == "piecewise":
            if None in (A, k_star, self.ng, self.nd):
                raise ValueError("Amplitude, k_star, n_g and n_d must be specified for piecewise shape")
            power_spectrum = np.piecewise(
                k,
                [k <= k_star, k > k_star],
                [lambda k: (k / k_star) ** self.ng,
                 lambda k: (k / k_star) ** -self.nd],
            )
            return A * np.maximum(power_spectrum, self.min_amplitude)

        elif self.shape == "lognormal":
            if None in (A, k_star, self.sigma_ln):
                raise ValueError("Amplitude, k_star and sigma_ln must be specified for lognormal shape")
            return A * np.exp(-0.5 * (np.log(k / k_star) / self.sigma_ln) ** 2)


        else:
            raise ValueError(f"Unknown power spectrum shape: {self.shape}")



import numpy as np
from scipy.interpolate import interp1d

class MassVariance:
    def __init__(self, window, power_spectrum, statistics=None, cutoff=False):
        """
        Initialize with a window function and a required power spectrum object.
        """
        self.window = window
        self.power_spectrum = power_spectrum
        self.statistics = statistics
        self.cutoff = cutoff

        if statistics == 'nonlinear' and self.window != "realtophat":
            raise ValueError('Must select real top hat window function for non-linear statistics')

    def window_function(self, k, R):
        """
        Computes the window function based on the chosen type ('gaussian', 'realtophat', etc.).
        """
        x = k * R

        if self.window == 'gaussian':
            return np.exp(-x**2 / 4)

        elif self.window == 'realtophat':
            if self.cutoff == True:
                result = 3 * (np.sin(x) - x * np.cos(x)) / x**3
                result = np.where(x > 4.49, 0.0, result)
                return result
            else:
                return 3 * (np.sin(x) - x * np.cos(x)) / x**3

        else:
            raise ValueError(f"Unknown window function: {self.window}")

    def evaluate(self, k_values):
        """
        Computes the integrand of the mass variance formula using explicit k_values.
        """
        kir_real = np.min(k_values)
        kuv_real = np.max(k_values)
        rir_real = 1 / kir_real
        rmin = 1 / kir_real
        rmax = 1 / kuv_real

        size = len(k_values)
        r = np.geomspace(rmin, rmax, size)
        k = k_values
        lnk = np.log(k)

        Pk = self.power_spectrum(k)

        sigma = {name: np.zeros_like(r) for name in [
            "sigma0", "sigma1peaks", "sigma1nl", "sigma2", "sigmaw",
            "sigmav", "sigmavg", "sigmagw", "sigmavw"]}

        pref = 16. / 81.

        for i in range(len(r)):
            kr = k * r[i]
            Wkr = self.window_function(k, r[i])
            Wkr2 = Wkr**2

            sigma["sigma0"][i] = pref * np.trapz(Wkr2 * kr**4 * Pk, lnk)
            sigma["sigma1peaks"][i] = pref * np.trapz(Wkr2 * kr**4 * k**2 * Pk, lnk)
            sigma["sigma1nl"][i] = pref * np.trapz(Wkr2 * kr**6 * Pk, lnk)
            sigma["sigma2"][i] = pref * np.trapz(Wkr2 * kr**8 * Pk, lnk)
            sigma["sigmaw"][i] = pref * np.trapz((Wkr * kr**4 - 2 * kr**2 * Wkr)**2 * Pk, lnk)
            sigma["sigmav"][i] = pref * np.trapz((3 * kr * np.sin(kr) - kr**2 * Wkr)**2 * Pk, lnk)
            sigma["sigmavg"][i] = pref * np.trapz((3 * kr * np.sin(kr) - kr**2 * Wkr) * Wkr * kr**2 * Pk, lnk)
            sigma["sigmagw"][i] = pref * np.trapz((kr**6 - 2 * kr**4) * Wkr2 * Pk, lnk)
            sigma["sigmavw"][i] = pref * np.trapz((3 * kr * np.sin(kr) - kr**2 * Wkr) * (Wkr * kr**4 - 2 * Wkr * kr**2) * Pk, lnk)

        sigma0f = interp1d(r, sigma["sigma0"], bounds_error=False, kind='cubic', fill_value="extrapolate")
        sigma1peaks = interp1d(r, sigma["sigma1peaks"], bounds_error=False, kind='cubic', fill_value="extrapolate")
        sigma1nl = interp1d(r, sigma["sigma1nl"], bounds_error=False, kind='cubic', fill_value="extrapolate")
        sigma2f = interp1d(r, sigma["sigma2"], bounds_error=False, kind='cubic', fill_value="extrapolate")
        sigmawf = interp1d(r, sigma["sigmaw"], bounds_error=False, kind='cubic', fill_value="extrapolate")
        sigmavf = interp1d(r, sigma["sigmav"], bounds_error=False, kind='cubic', fill_value="extrapolate")
        sigmavgf = interp1d(r, sigma["sigmavg"], bounds_error=False, kind='cubic', fill_value="extrapolate")
        sigmagwf = interp1d(r, sigma["sigmagw"], bounds_error=False, kind='cubic', fill_value="extrapolate")
        sigmavwf = interp1d(r, sigma["sigmavw"], bounds_error=False, kind='cubic', fill_value="extrapolate")

        if self.statistics == 'press':
            return sigma0f

        elif self.statistics == 'peaks':
            return sigma0f, sigma1nl

        elif self.statistics == 'nonlinear':
            return sigma0f, sigma1nl, sigma2f, sigmavf, sigmawf, sigmavgf, sigmagwf, sigmavwf

        else:
            raise ValueError(f"Unknown statistics option: {self.statistics}")



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import LogLocator, FormatStrFormatter
import pickle

plt.rcParams.update({
"text.usetex": False,
"font.family": "serif",
"font.sans-serif": ["TimesNewRoman"]})
plt.rcParams["mathtext.fontset"] = "stix"

# Helper functions for plotting main quantities

def plot_power_spectrum(k_vals, p_k):

    # Create figure and axis
    fig, ax = plt.subplots(figsize=(8, 6))

    # Plot the custom power spectrum
    ax.loglog(k_vals, p_k, label=r"$\mathcal{P}_\zeta(k)$", color="navy", lw=2)

    # Axis labels and title
    ax.set_xlabel(r"$k$ [$\mathrm{Mpc}^{-1}$]", fontsize=14)
    ax.set_ylabel(r"$\mathcal{P}_\zeta(k)$", fontsize=14)
    ax.set_title("Initial Power Spectrum", fontsize=14)

    # Ticks
    ax.tick_params(axis='both', which='major', labelsize=14)
    ax.xaxis.set_major_locator(LogLocator(base=10.0, numticks=10))
    ax.yaxis.set_major_locator(LogLocator(base=10.0, numticks=10))
    ax.xaxis.set_minor_formatter(FormatStrFormatter(""))

    # Layout
    plt.tight_layout()
    plt.savefig('powerspectrum.pdf')
    plt.show()

def plot_mass_variance(r_vals, sigma):
    if isinstance(sigma, tuple):
        sigma_vals = sigma[0](r_vals)
    else: 
        sigma_vals = sigma(r_vals)


    # Create figure and axis
    fig, ax = plt.subplots(figsize=(8, 6))

    # Plot the mass variance
    ax.loglog(r_vals, sigma_vals, color='darkgreen', lw=2)

    # Axis labels and title
    ax.set_xlabel(r"$R$ [$\mathrm{Mpc}$]", fontsize=14)
    ax.set_ylabel(r"$\sigma^2(R)$", fontsize=14)
    ax.set_title("Mass Variance", fontsize=14)

    # Ticks
    ax.xaxis.set_major_locator(LogLocator(base=10.0, numticks=10))
    ax.yaxis.set_major_locator(LogLocator(base=10.0, numticks=10))
    ax.xaxis.set_minor_formatter(FormatStrFormatter(""))

    # Layout
    plt.tight_layout()
    plt.savefig('massvariance.pdf')
    plt.show()

def plot_mass_function(m_vals, f_m_vals, fpbh_val = None):
    fig, ax = plt.subplots(figsize=(8, 6))

    # Plot the mass function
    ax.loglog(m_vals, f_m_vals, label=r"$f_{\rm PBH}(M)$", color="crimson", lw=2)

    # Set limits
    #ax.set_ylim(1e-5, 1)

    # Axis labels and title
    ax.set_xlabel(r"$M_{\rm PBH}$ [$M_\odot$]", fontsize=14)
    ax.set_ylabel(r"$f\,(M)$", fontsize=14)
    ax.set_title("Primordial Black Hole Mass Function", fontsize=14)

    # Ticks
    ax.xaxis.set_major_locator(LogLocator(base=10.0, numticks=10))
    ax.yaxis.set_major_locator(LogLocator(base=10.0, numticks=10))
    ax.xaxis.set_minor_formatter(FormatStrFormatter(""))
    
    if fpbh_val is not None:
        ax.text(m_vals.min(), f_m_vals.max(), rf"$f_{{\rm PBH}} = {fpbh_val:.2e}$", fontsize=12, color="black",
                bbox=dict(facecolor='white', alpha=0.8, edgecolor='gray'))
    
    plt.tight_layout()
    plt.savefig('massfunction.pdf')
    plt.show()

def bounds_utility(m_vals, f_m_vals, path_to_pbhbounds):
    np.savetxt(path_to_pbhbounds + 'm_array.txt', m_vals)
    np.savetxt(path_to_pbhbounds + 'f_array.txt', f_m_vals)
    return None



import numpy as np
from math import erfc
from scipy.interpolate import interp1d
from scipy.special import hyp2f1

vec_erfc = np.vectorize(erfc)

# Error function variant
def Erf(x):
    return -1 + vec_erfc(-x)

# Initialize gcf
def initialize_gcf(wmax):
    qq = np.geomspace(0.0001, wmax, 10000)

    def th(q):
        hyp = hyp2f1(1, 5 / (2 + 2 * q), 1 + 5 / (2 + 2 * q), -1 / q)
        return np.clip(2 * q / ((3 * (1 + q)) * hyp), 0, 1)

    th_vals = th(qq)
    def wq(q, th_vals):
        return q * th_vals * 4 * np.sqrt(1 - 3 / 2 * th_vals)

    wq_vals = wq(qq, th_vals)

    wqf = interp1d(qq, wq_vals, kind='cubic', bounds_error=False, fill_value="extrapolate")
    qw = interp1d(wq_vals, qq, kind='cubic', bounds_error=False, fill_value="extrapolate")

    ww = np.geomspace(0.0001, wmax, 10000)

    def gc(w, qw):
        q_vals = qw(w)
        th_vals = th(q_vals)
        y = 4/3 * (1 - np.sqrt(1 - 3/2 * th_vals))
        y2 = 4/3 - 32/9 * 1/w
        return np.where(w < 1e4, y, y2)

    gca = gc(ww, qw)
    gcf0 = interp1d(ww, np.log(gca), kind='cubic', bounds_error=False, fill_value="extrapolate")

    def gcf(w):
        return np.exp(gcf0(w))

    return gcf

# fsm and fs
def fsm(x):
    if x > 0.1:
        y = (x**3 - 3 * x)/2 * (Erf(np.sqrt(5/2) * x) + Erf(np.sqrt(5/2) * x/2)) + \
            np.sqrt(2 / (5 * np.pi)) * ((31 * x**2 / 4 + 8/5) * np.exp(-5 * x**2 / 8) + \
                                        (x**2 / 2 - 8/5) * np.exp(-5 * x**2 / 2))
    else:
        xsmall = x
        x = 0.1
        y = (x**3 - 3 * x)/2 * (Erf(np.sqrt(5/2) * x) + Erf(np.sqrt(5/2) * x/2)) + \
            np.sqrt(2 / (5 * np.pi)) * ((31 * x**2 / 4 + 8/5) * np.exp(-5 * x**2 / 8) + \
                                        (x**2 / 2 - 8/5) * np.exp(-5 * x**2 / 2))
        y = (xsmall / 0.1)**8 * y
    return y

x = np.geomspace(0.0001, 100, 1000)
fs0 = np.array([fsm(xi) for xi in x])
fs = interp1d(x, fs0, kind='cubic', bounds_error=False, fill_value="extrapolate")

# cg relations
def C(g):
    return g * (1 - 3/8 * g)

def crit(g, w, gam=0.36):
    return (C(g) - C(gcf(w)))**gam

def gamma_f(r, sigmagwf, sigmagf, sigmawf):
    return sigmagwf(r) / np.sqrt(sigmagf(r) * sigmawf(r))

def barg(r, sigmagwf, sigmawf):
    return sigmagwf(r) / sigmawf(r)

def sigmawtildef(r, sigmavf, sigmawf, sigmavwf):
    return (sigmavf(r) * sigmawf(r) - sigmavwf(r)**2) / sigmavf(r)

def bargfull(r, sigmavf, sigmagwf, sigmavwf, sigmavgf, sigmawf):
    return (sigmavf(r) * sigmagwf(r) - sigmavwf(r) * sigmavgf(r)) / (sigmavf(r) * sigmawf(r) - sigmavwf(r)**2)

def Sigmagf(r, sigmagf, sigmawf, sigmavf, sigmavgf, sigmagwf, sigmavwf):
    numerator = (sigmavf(r) * sigmagf(r) * sigmawf(r)
                 - sigmavgf(r)**2 * sigmawf(r)
                 - sigmavwf(r)**2 * sigmagf(r)
                 - sigmagwf(r)**2 * sigmavf(r)
                 + 2 * sigmagwf(r) * sigmavwf(r) * sigmavgf(r))
    denominator = sigmavf(r) * sigmawf(r) - sigmavwf(r)**2
    return numerator / denominator

def Mmax(r, alpha, A, K, gamp):
    wm1 = 2/3 * 1/np.log(alpha) * alpha**2 * r**2
    wm = (wm1 + np.sqrt(wm1**2 + 4 * 2.28 * A * 2/9 * alpha**4 * r**4)) / 2
    M = K * r**2 * (2**7 / 3**3)**gamp / wm**(2*gamp) 
    return M

def find_peak_range_indices(ps, threshold_factor=0.1):
    """
    Find indices where power spectrum is within one order of magnitude from the peak.
    
    Parameters:
        ps (array): Power spectrum values.
        threshold_factor (float): Fraction of the peak to use (default 0.1 = one order of magnitude).
        
    Returns:
        idx_min (int): First index in the peak region.
        idx_max (int): Last index in the peak region.
        peak_indices (ndarray): Array of all indices within the peak range.
    """
    ps = np.asarray(ps)
    ps_max = np.max(ps)
    mask = ps >= threshold_factor * ps_max
    peak_indices = np.where(mask)[0]
    
    if peak_indices.size == 0:
        raise ValueError("No peak region found above threshold.")
    
    idx_min = peak_indices[0]
    idx_max = peak_indices[-1]
    
    return idx_min, idx_max


from distutils.core import setup
setup(
  name = 'pbhstat',
  packages = ['pbhstat'],
  version = '0.0.1',
  license='MIT',
  description = 'A Python code for calculating the primordial black hole mass function and abundance.',
  author = 'Philippa Cole and Jacopo Fumagalli',
  author_email = 'philippa.cole@unimib.it',
  url = 'https://github.com/pipcole/pbhstat',
  download_url = 'https://github.com/pipcole/pbhstat/archive/refs/tags/v.0.0.1.tar.gz',
  keywords = ['PBH', 'abundance', 'mass function'],
  install_requires=[            # I get to this in a second
          'numpy',
          'scipy',
          'matplotlib'
          'tqdm',
      ],
  classifiers=[
    'Development Status :: 4 - Beta',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package

    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Software Development :: Build Tools',

    'License :: OSI Approved :: MIT License',   # Again, pick a license

    'Programming Language :: Python :: 3.9',      #Specify which pyhton versions that you want to support
    'Programming Language :: Python :: 3.11',
    'Programming Language :: Python :: 3.12',
  ],
)


import numpy as np
from matplotlib import pyplot as plt
import tqdm

from constants import OmegaCDM, Meq, req, geq

from stats_utils import (
    C, fs, initialize_gcf, gamma_f, barg, sigmawtildef, bargfull,
    Sigmagf, Mmax, find_peak_range_indices
)

class PressSchechterModel:
    def __init__(self, power_spectrum, mass_variance, K, C_c, gamma=0.36):
        self.power_spectrum = power_spectrum
        self.mass_variance = mass_variance
        self.K = K
        self.gamma = gamma
        self.C_c = C_c

        if self.mass_variance.window == 'gaussian' and (K != 10 or C_c != 0.28):
            print('Warning: Gaussian window function typically uses K=10 and C_c=0.28')
        if self.mass_variance.window == 'realtophat' and (K != 4 or C_c != 0.77):
            print('Warning: Top hat window function typically uses K=4 and C_c=0.77')

    def mu(self, mh, mbh):
        return mbh / (self.K * mh)

    def C(self, mh, mbh):
        return self.mu(mh, mbh) ** (1 / self.gamma) + self.C_c

    def P(self, C_val, sigma_sq):
        return (1 / np.sqrt(2 * np.pi * sigma_sq)) * np.exp(-C_val**2 / (2 * sigma_sq))

    def beta_mh_integrand(self, mh, mbh, sigma_func):
        mu_val = self.mu(mh, mbh)
        C_val = self.C(mh, mbh)
        R = np.sqrt(mh / Meq) * req  # Inverting MH = Meq * (R/req)^2
        sigma_sq = sigma_func(R)
        prefactor = (1 / OmegaCDM) * np.sqrt(Meq / mh) # prev 1/2^(5/2) at front
        return prefactor * mu_val**(1 / self.gamma) * (mbh / (self.gamma * mh**2)) * self.P(C_val, sigma_sq)

    def evaluate(self, k_values, mpbh_vals):

        klow, khigh = min(k_values), max(k_values)
        rlow, rhigh = 1 / khigh, 1 / klow

        # Arrays for integration
        r_array = np.geomspace(rlow, rhigh, 3000)
        MH_array = Meq * (r_array / req)**2

        sigma_func = self.mass_variance.evaluate(k_values)  # interpolator σ2(R)

        MPBH = np.geomspace(min(MH_array) * 10, max(MH_array), mpbh_vals)
        fm_vals = np.zeros_like(MPBH)

        for i, mbh in enumerate(MPBH):
            integrand = self.beta_mh_integrand(MH_array, mbh, sigma_func)
            fm_vals[i] = np.trapz(integrand, MH_array)

        return MPBH, fm_vals

class PeaksTheoryModel:
    def __init__(self, power_spectrum, mass_variance, K, C_c, gamma=0.36):
        self.power_spectrum = power_spectrum
        self.mass_variance = mass_variance
        self.K = K
        self.gamma = gamma
        self.C_c = C_c

        if mass_variance.window == 'gaussian' and (K != 10 or C_c != 0.28):
            print('Warning: Gaussian window function usually takes K=10, C_c=0.28')
        if self.mass_variance.window == 'realtophat' and (K != 4 or C_c != 0.77):
            print('Warning: Top hat window function typically uses K=4 and C_c=0.77')

    def evaluate(self, k_values, mpbh_vals):
        sigma0f, sigma1f = self.mass_variance.evaluate(k_values)

        k_low = min(k_values)
        k_high = max(k_values)
        r_low = 1 / k_high
        r_high = 1 / k_low

        r_array = np.geomspace(r_low, r_high, 3000)
        MH_array = Meq * (r_array / req)**2

        MPBH = np.geomspace(min(MH_array) * 10, max(MH_array), mpbh_vals)

        def mu(mh, mbh):
            return mbh / (self.K * mh)

        def C(mh, mbh):
            return mu(mh, mbh)**(1 / self.gamma) + self.C_c

        def nu(mh, mbh, sig0):
            R = np.sqrt(mh / Meq) * req
            return C(mh, mbh) / np.sqrt(sig0(R))

        def npeaks(mh, mbh, sig0, sig1):
            R = np.sqrt(mh / Meq) * req 
            ratio = np.sqrt(sig1(R) / sig0(R))
            n = nu(mh, mbh, sig0)
            return (1 / (3**1.5 * (2 * np.pi)**2)) * ratio**3 * n**3 * np.exp(-0.5 * n**2)

        def beta_integrand_old(mh, mbh, sig0, sig1):
            R = np.sqrt(mh / Meq) * req 
            mu_val = mu(mh, mbh)
            n_peak = npeaks(mh, mbh, sig0, sig1)
            if self.mass_variance.window == 'gaussian':
                b_pref = (2 * np.pi)**1.5
            else: b_pref = 4 * np.pi/3
            prefactor = (1/np.sqrt(2)) * (1 / OmegaCDM) * b_pref
            return prefactor * (req**3 / Meq) * (mbh / (self.gamma * mh)) * mu_val**(1 / self.gamma) * n_peak

        def beta_integrand(mh, mbh, sig0, sig1):
            R = np.sqrt(mh / Meq) * req 
            mu_val = mu(mh, mbh)
            n_peak = npeaks(mh, mbh, sig0, sig1)
            if self.mass_variance.window == 'gaussian':
                b_pref = (2 * np.pi)**1.5
            else: b_pref = 4 * np.pi/3
            prefactor = (1/2) * (1 / OmegaCDM) * b_pref
            return prefactor * np.sqrt(Meq/mh**3) * (mbh / (self.gamma * mh)) * mu_val**(1 / self.gamma) * n_peak
        
        fm_vals = np.zeros_like(MPBH)
        for i, mbh in enumerate(MPBH):
            integrand_vals = beta_integrand(MH_array, mbh, sigma0f, sigma1f)
            fm_vals[i] = np.trapz(integrand_vals, MH_array)

        return MPBH, fm_vals

class NonLinearModel:
    def __init__(self, power_spectrum, mass_variance, K, gamma=0.36, vcorr=True):
        self.power_spectrum = power_spectrum
        self.mass_variance = mass_variance
        self.K = K
        self.gamma = gamma
        self.vcorr = vcorr

        if mass_variance.window == 'gaussian' and (K != 10):
            print('Warning: Gaussian window function usually takes K=10')
        if self.mass_variance.window == 'realtophat' and (K != 4):
            print('Warning: Top hat window function typically uses K=4')

    def evaluate(self, k_values, mpbh_vals):
        ps = self.power_spectrum(k_values)
        sigma0f, sigma1f, sigma2f, sigmavf, sigmawf, sigmavgf, sigmagwf, sigmavwf = self.mass_variance.evaluate(k_values)

        k_low = min(k_values)
        k_high = max(k_values)

        idx_min, idx_max = find_peak_range_indices(ps, threshold_factor=0.01)

        kir_real = k_values[idx_min]
        kuv_real = k_values[idx_max]
        alpha = kuv_real / kir_real

        rir_real = 1 / k_low
        ruv_real = 1 / kuv_real

        r_array = np.geomspace(ruv_real, rir_real, 3000)
        MH_array = Meq * (r_array / req)**2 
        LMH = np.log(MH_array)

        MPBH_array = np.geomspace(min(MH_array), max(MH_array), mpbh_vals)

        if alpha < 1.3:
            wmin = 0.1
            wmax =  (rir_real*kuv_real)**2 / (2 * np.log(alpha)) * 4/3 * 50
        else: # require larger wmax and smaller wmin for broad spectra
            wmin = 0.01
            wmax =  (rir_real*kuv_real)**2 / (2 * np.log(alpha)) * 4/3 * 1000 

        w_array = np.geomspace(wmin, wmax, 4000)

        gcf = initialize_gcf(wmax)

        def integrand_w(w, r, M, MH, vcorr=True):
            muj = M / (self.K * MH)
            pM = M / MH**(3/2)
            pv = 1 / np.sqrt(2 * np.pi * sigmavf(r))
            condition = 1 - 3/2 * C(gcf(w)) - 3/2 * muj**(1 / self.gamma)

            g = 4/3 * (1 - np.sqrt(np.abs(condition)))
            dgoverdlogM = 1 / (1 - 3/4 * g) * 1 / self.gamma * muj**(1 / self.gamma)

            ronrstar = (1/3 * sigma2f(r) / sigma1f(r))**(3/2)
            prefactor = 2/3 * np.pi / (2 * np.pi)**(3/2)

            factor_1 = prefactor * ronrstar * dgoverdlogM * fs((2 * g + w) / np.sqrt(sigma2f(r))) * pM * pv

            if self.vcorr:
                bg = bargfull(r, sigmavf, sigmagwf, sigmavwf, sigmavgf, sigmawf)
                sg = Sigmagf(r, sigma0f, sigmawf, sigmavf, sigmavgf, sigmagwf, sigmavwf)
                sw = sigmawtildef(r, sigmavf, sigmawf, sigmavwf)
                pg = np.exp(-(g - bg * w)**2 / (2 * sg)) / np.sqrt(2 * np.pi * sg)
                pw = np.exp(-w**2 / (2 * sw)) / np.sqrt(2 * np.pi * sw)
            else:
                bg = barg(r, sigmagwf, sigmawf)
                sg = sigma0f(r)
                sw = sigmawf(r)
                gam = gamma_f(r, sigmagwf, sigma0f, sigmawf)
                pg = np.exp(-(g - bg * w)**2 / (2 * sg * (1 - gam**2))) / np.sqrt(2 * np.pi * sg * np.sqrt((1 - gam**2)**2))
                pw = np.exp(-w**2 / (2 * sw)) / np.sqrt(2 * np.pi * sw)

            return np.where(condition > 0, factor_1 * w * pg * pw, 0)

        integrand_mh = np.zeros([len(MH_array), len(MPBH_array)])
        fm_vals = np.zeros_like(MPBH_array)

        for j in tqdm.tqdm(range(len(MPBH_array)), desc="Evaluating NonLinearModel"):
            for i in range(len(MH_array)):
                integrand_mh[i, j] = np.trapz(integrand_w(w_array, r_array[i], MPBH_array[j], MH_array[i], self.vcorr), w_array) + 1e-70
            fm_vals[j] = np.trapz(integrand_mh[:, j], LMH)

        return MPBH_array, fm_vals * np.sqrt(Meq)
